/*
 * nsVanadielAlarm.js
 *
 * Copyright (c) 2005 by Daniel Potter.  All rights reserved.
 *
 * This file is part of the Vana'diel Clock.
 *
 * The Vana'diel Clock is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The Vana'diel Clock is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Vana'diel Clock; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Manages alarms.
 */

const VANADIEL_CLOCK_NAMESPACE = "http://ffxi.xenoveritas.org/tools/vanaclock/rdf#";
const VERSION_PROPERTY = VANADIEL_CLOCK_NAMESPACE + "version";
const ALARM_PROPERTY = VANADIEL_CLOCK_NAMESPACE + "alarm";
const USER_ALARM_PREFIX = VANADIEL_CLOCK_NAMESPACE + "user-alarm-";
const ALARM_PREFIX = "http://ffxi.xenoveritas.org/tools/vanaclock/rdf#alarm-";
const ALARM_NAME_PROPERTY = ALARM_PREFIX + "name";
const ALARM_DESCRIPTION_PROPERTY = ALARM_PREFIX + "description";
const ALARM_ALARMS_PROPERTY = ALARM_PREFIX + "alarms";
const ALARM_PERIOD_PROPERTY = ALARM_PREFIX + "period";
const ALARM_ICON_PROPERTY = ALARM_PREFIX + "icon";
const ALARM_SMALL_ICON_PROPERTY = ALARM_PREFIX + "icon-small";
const ALARM_SOUND_PROPERTY = ALARM_PREFIX + "sound";
const ALARM_TYPE_PROPERTY = ALARM_PREFIX + "type";

var lastAlarmGUID = 0;

function VanadielAlarm() {
    this.mAlarms = new Array();
    var stamp = Math.floor(new Date().getTime() / 1000);
    if (stamp == lastAlarmGUID) {
	stamp++;
    }
    lastAlarmGUID = stamp;
    this.mId = "A" + stamp;
}

VanadielAlarm.prototype = {
    RDF: Components.classes["@mozilla.org/rdf/rdf-service;1"]
	       .getService(Components.interfaces.nsIRDFService),
    mId: null,
    mNextAlarm: 0,
    mAlarms: null,
    mPeriod: 0,
    mName: "",
    mDescription: "",
    mIcon: "",
    mSmallIcon: "",
    mSound: "",
    mType: "",
    mRDFSource: null,
    mRDFNode: null,
    
    get id() { return this.mId; },
    
    get nextAlarm() { return this.mNextAlarm; },
    
    get active() {
	return this.mAlarms.length > 0;
    },

    get period() { return this.mPeriod; },
    set period(value) {
	if (value >= 0) {
	    this.mPeriod = value;
	    this.setRDF(ALARM_PERIOD_PROPERTY, value);
	}
    },

    get name() { return this.mName; },
    set name(value) {
	if (value != null) {
	    this.mName = value;
	    this.setRDF(ALARM_NAME_PROPERTY, value);
	}
    },

    get description() { return this.mDescription; },
    set description(value) {
	if (value != null) {
	    this.mDescription = value;
	    this.setRDF(ALARM_DESCRIPTION_PROPERTY, value);
	}
    },

    get icon() { return this.mIcon; },
    set icon(value) {
	if (value != null) {
	    this.mIcon = value;
	    this.setRDF(ALARM_ICON_PROPERTY, value);
	}
    },

    get smallIcon() { return this.mSmallIcon; },
    set smallIcon(value) {
	if (value != null) {
	    this.mSmallIcon = value;
	    this.setRDF(ALARM_SMALL_ICON_PROPERTY, value);
	}
    },

    get sound() { return this.mSound; },
    set sound(value) {
	if (value != null) {
	    this.mSound = value;
	    this.setRDF(ALARM_SOUND_PROPERTY, value);
	}
    },
    
    get type() { return this.mType; },
    set type(value) {
	if (value != null) {
	    this.mType = value;
	    this.setRDF(ALARM_TYPE_PROPERTY, value);
	}
    },

    setAlarm: function(time) {
	this.mAlarms = new Array(1);
	this.mAlarms[0] = time;
	this.mNextAlarm = time;
    },
    
    getAlarms: function(count) {
	if (count < 0 || count >= this.mAlarms.length)
	    return 0;
	return this.mAlarms[count];
    },

    addAlarm: function(time) {
	if (this.mAlarms.length == 0) {
	    this.mNextAlarm = time;
	} else {
	    if (time < this.mNextAlarm) {
		this.mNextAlarm = time;
	    }
	}
	this.mAlarms.push(time);
	this.setRDF(ALARM_ALARMS_PROPERTY, this.mAlarms.join(","));
    },
    
    activated: function(time) {
	var a = 0;
	while (a < this.mAlarms.length) {
	    if (this.mAlarms[a] <= time) {
		// if we're set to reactive, reactivate
		if (this.mPeriod > 0) {
		    var periods = Math.floor((time - this.mAlarms[a]) / this.mPeriod);
		    periods++;
		    this.mAlarms.push(this.mAlarms[a] + this.mPeriod * periods);
		}
		// remove this time
		this.mAlarms.splice(a, 1);
	    } else {
		if (this.mAlarms[a] > this.mNextAlarm)
		    this.mNextAlarm = this.mAlarms[a];
		a++;
	    }
	}
	this.setRDF(ALARM_ALARMS_PROPERTY, this.mAlarms.join(","));
	return this.mAlarms.length > 0;
    },
    
    /**
     * Loads the information from the given rdfResource.
     * This may reset the autogenerated ID.
     */
    bindToRDF: function(rdfSource, rdfResource) {
	this.mRDFSource = rdfSource;
	this.mRDFNode = rdfResource;
	if (rdfResource.Value.substring(0, USER_ALARM_PREFIX.length) == USER_ALARM_PREFIX) {
	    dump("ID changed from ");
	    dump(this.mId);
	    this.mId = rdfResource.Value.substring(USER_ALARM_PREFIX.length);
	    dump(" to ");
	    dump(this.mId);
	    dump(".\n");
	}
	this.mName = this.getOrSetRDF(ALARM_NAME_PROPERTY, this.mName);
	this.mDescription = this.getOrSetRDF(ALARM_DESCRIPTION_PROPERTY, this.mDescription);
	this.mIcon = this.getOrSetRDF(ALARM_ICON_PROPERTY, this.mIcon);
	this.mSmallIcon = this.getOrSetRDF(ALARM_SMALL_ICON_PROPERTY, this.mSmallIcon);
	this.mSound = this.getOrSetRDF(ALARM_SOUND_PROPERTY, this.mSound);
	this.mType = this.getOrSetRDF(ALARM_TYPE_PROPERTY, this.mType);
	var alarms = this.getOrSetRDF(ALARM_ALARMS_PROPERTY, this.mAlarms.join(",")).split(",");
	this.mPeriod = parseInt(this.getOrSetRDF(ALARM_PERIOD_PROPERTY, this.mPeriod));
	if (! (this.mPeriod >= 0)) {
	    // corrects errors (since NaN is always false)
	    this.mPeriod = 0;
	}
	this.mAlarms = new Array();
	this.mNextAlarm = NaN;
	for (var i in alarms) {
	    var t = parseInt(alarms[i]);
	    if (t > 0) { // prevent errors from being added
		this.mAlarms.push(t);
		if (!(this.mNextAlarm > t))
		    this.mNextAlarm = t;
	    }
	}
	this.dumpAlarm();
    },
    
    unbindFromRDF: function() {
	if (this.mRDFSource == null)
	    return;
	// unassert everything
	this.unsetRDF(ALARM_NAME_PROPERTY);
	this.unsetRDF(ALARM_DESCRIPTION_PROPERTY);
	this.unsetRDF(ALARM_ALARMS_PROPERTY);
	this.unsetRDF(ALARM_PERIOD_PROPERTY);
	this.unsetRDF(ALARM_ICON_PROPERTY);
	this.unsetRDF(ALARM_SMALL_ICON_PROPERTY);
	this.unsetRDF(ALARM_SOUND_PROPERTY);
	this.unsetRDF(ALARM_TYPE_PROPERTY);
	this.mRDFSource = null;
	this.mRDFNode = null;
    },
    
    getOrSetRDF: function(name, value) {
	var res = this.RDF.GetResource(name);
	var v = this.mRDFSource.GetTarget(this.mRDFNode, res, true);
	if (v == null) {
	    // set it
	    this.mRDFSource.Assert(this.mRDFNode, res, this.RDF.GetLiteral(value), true);
	    return value;
	} else {
	    return v.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
	}
    },

    setRDF: function(name, value) {
	// allow sets if we aren't bound...
	if (this.mRDFSource == null)
	    return;
	var res = this.RDF.GetResource(name);
	var v = this.mRDFSource.GetTarget(this.mRDFNode, res, true);
	if (v == null) {
	    // set it
	    this.mRDFSource.Assert(this.mRDFNode, res, this.RDF.GetLiteral(value), true);
	} else {
	    // change it
	    this.mRDFSource.Change(this.mRDFNode, res, v, this.RDF.GetLiteral(value));
	}
    },
    
    unsetRDF: function(name) {
	var res = this.RDF.GetResource(name);
	var v = this.mRDFSource.GetTarget(this.mRDFNode, res, true);
	if (v != null) {
	    // unassert it
	    this.mRDFSource.Unassert(this.mRDFNode, res, v);
	}
    },

    QueryInterface: function (iid) {
        if (!iid.equals(Components.interfaces.nsIVanadielAlarm)
            && !iid.equals(Components.interfaces.nsISupports))
        {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    },
    
    dumpAlarm: function () {
	dump("Alarm:\n");
	dump("            id: " + this.mId + "\n");
	dump("     nextAlarm: " + this.mNextAlarm + "\n");
	dump("        alarms: " + this.mAlarms + "\n");
	dump("        period: " + this.mPeriod + "\n");
	dump("          name: " + this.mName + "\n");
	dump("   description: " + this.mDescription + "\n");
	dump("          icon: " + this.mIcon + "\n");
	dump("     smallIcon: " + this.mSmallIcon + "\n");
	dump("         sound: " + this.mSound + "\n");
	dump("          type: " + this.mType + "\n");
    }
}

function VanadielAlarmManager() {
    this.alarms = new Array();
    var profileInternal = Components.classes['@mozilla.org/file/directory_service;1']
                                    .getService(Components.interfaces.nsIProperties);
    var file = profileInternal.get('ProfD', Components.interfaces.nsILocalFile);
    file.append("vanadiel-alarms.rdf");
    var fileHandler = Components.classes["@mozilla.org/network/io-service;1"]
                        .getService(Components.interfaces.nsIIOService);
    var url = fileHandler.newFileURI(file);
    this.mRdfUri = url.spec;
    dump("loading alarms from " + this.mRdfUri + "\n");
    try {
	this.alarmDS = this.RDF.GetDataSourceBlocking(this.mRdfUri);
	// see if there's a version already set
	this.clockNS = this.RDF.GetResource(VANADIEL_CLOCK_NAMESPACE);
	this.alarmNS = this.RDF.GetResource(ALARM_PROPERTY);
	var versionNS = this.RDF.GetResource(VERSION_PROPERTY);
	var version = this.alarmDS.GetTarget(this.clockNS, versionNS, true);
	if (version == null) {
	    // No - set the version
	    this.alarmDS.Assert(this.clockNS, versionNS, this.RDF.GetLiteral("1.0"), true);
	}
	// Try and get the alarm RDF sequence
	if (this.RDFCUtils.IsSeq(this.alarmDS, this.alarmNS)) {
	    this.alarmRDFSeq = Components.classes['@mozilla.org/rdf/container;1']
				    .createInstance(Components.interfaces.nsIRDFContainer);
	    this.alarmRDFSeq.Init(this.alarmDS, this.alarmNS);
	} else {
	    // create it
	    this.alarmRDFSeq = this.RDFCUtils.MakeSeq(this.alarmDS, this.alarmNS);
	}
	this.alarmDS.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
    } catch (ex) {
	// TODO: send JS errors to the JavaScript console?
	dump("Failed to load alarm from RDF:\n");
	dump(ex);
	dump("\n");
	for (var key in ex) {
	    dump("exception.");
	    dump(key);
	    dump(": ");
	    dump(ex[key]);
	    dump("\n");
	}
    }
    this.mListeners = new Array();
}

VanadielAlarmManager.prototype = {
    alarms: null,
    RDF: Components.classes["@mozilla.org/rdf/rdf-service;1"]
	    .getService(Components.interfaces.nsIRDFService),
    RDFCUtils: Components.classes["@mozilla.org/rdf/container-utils;1"]
		.getService(Components.interfaces.nsIRDFContainerUtils),
    alarmDS: null,
    clockNS: null,
    alarmNS: null,
    alarmRDFSeq: null,
    loaded: false,
    mRdfUri: "",
    mListeners: null,
    
    get rdfURI() { return this.mRdfUri; },
    
    get count() { return this.alarms.length; },
    
    getAlarmAt: function(index) {
	if (index < 0 || index >= this.alarms.length)
	    return null;
	return this.alarms[index];
    },
    
    getAlarmById: function(id) {
	for (var a in this.alarms) {
	    if (this.alarms[a].id == id)
		return this.alarms[a];
	}
	return null;
    },
    
    addAlarm: function(alarm) {
	this.alarms.push(alarm);
	// add the alarm to our datasource (creating a GUID for this alarm)
	var alarmNode = this.RDF.GetResource(USER_ALARM_PREFIX + alarm.id);
	this.alarmRDFSeq.AppendElement(alarmNode);
	alarm.bindToRDF(this.alarmDS, alarmNode);
	this.flushAlarms();
	// alert listeners that an alarm has come into existance
	for (var l in this.mListeners) {
	    try {
		this.mListeners[l].alarmAdded(alarm);
	    } catch (ex) {
		// alert on error
		// TODO: send JS errors to the JavaScript console?
		dump("Listener caused exception:\n");
		dump(ex);
		dump("\n");
		for (var key in ex) {
		    dump("exception.");
		    dump(key);
		    dump(": ");
		    dump(ex[key]);
		    dump("\n");
		}
	    }
	}
    },
    
    removeAlarm: function(alarm) {
	// Find this alarm in the array
	var index;
	for (index = 0; index < this.alarms.length; index++) {
	    if (this.alarms[index] == alarm) {
		// remove the alarm
		var node = this.RDF.GetResource(USER_ALARM_PREFIX + alarm.id);
		this.alarmRDFSeq.RemoveElement(node, true);
		alarm.unbindFromRDF();
		// remove the alarm from the array
		this.alarms.splice(index, 1);
		this.flushAlarms();
		// alert listeners that alarms have vanished
		for (var l in this.mListeners) {
		    try {
			this.mListeners[l].alarmRemoved(alarm);
		    } catch (ex) {
			// alert on error
			// TODO: send JS errors to the JavaScript console?
			dump("Listener caused exception:\n");
			dump(ex);
			dump("\n");
			for (var key in ex) {
			    dump("exception.");
			    dump(key);
			    dump(": ");
			    dump(ex[key]);
			    dump("\n");
			}
		    }
		}
		break;
	    }
	}
    },
    
    loadAlarms: function() {
	if (this.loaded)
	    return;
	this.loaded = true;
	var elements = this.alarmRDFSeq.GetElements();
	this.alarms = new Array();
	// go through the alarms and create them
	while (elements.hasMoreElements()){
	    dump("Collecting alarm\n");
	    var alarmRes = elements.getNext();
	    if (alarmRes instanceof Components.interfaces.nsIRDFResource) {
		// create this alarm
		var alarm = Components.classes["@xenoveritas.org/vanadiel-alarm;1"].
			   createInstance(Components.interfaces.nsIVanadielAlarm);
		try {
		    alarm.bindToRDF(this.alarmDS, alarmRes);
		    this.alarms.push(alarm);
		} catch (ex) {
		    // TODO: send JS errors to the JavaScript console?
		    dump("Error loading alarm \"");
		    dump(alarmRes.Value);
		    dump("\":\n");
		    dump(ex);
		    dump("\n");
		    for (var key in ex) {
			dump("exception.");
			dump(key);
			dump(": ");
			dump(ex[key]);
			dump("\n");
		    }
		}
	    }
	}
    },
    
    flushAlarms: function() {
	this.alarmDS.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
    },
    
    activateAlarms: function(time, late) {
	var alarmTitle = null;
	var alarmDesc = null;
	var alarmIcon = null;
	var sound = null;
	for (var a in this.alarms) {
	    if (this.alarms[a].nextAlarm <= time) {
		this.alarms[a].activated(time);
		if (alarmTitle == null) {
		    alarmTitle = this.alarms[a].name;
		    alarmDesc = this.alarms[a].description;
		    alarmSound = this.alarms[a].sound;
		    alarmIcon = this.alarms[a].icon;
		} else {
		    alarmTitle = alarmTitle + ", " + this.alarms[a].name;
		    alarmDesc = alarmDesc + ", " + this.alarms[a].description;
		    // "upgrade" the alarm sound, if possible
		    if (alarmSound == "none") {
			alarmSound = this.alarms[a].sound;
		    } else if (alarmSound == "beep") {
			if (this.alarms[a].sound != "none") {
			    alarmSound = this.alarms[a].sound;
			}
		    }
		}
	    }
	}
	if (alarmTitle != null) {
	    var alertService = Components.classes["@mozilla.org/alerts-service;1"].getService(Components.interfaces.nsIAlertsService);
	    alertService.showAlertNotification(alarmIcon, alarmTitle, alarmDesc, false, "don't care", null);
	    if (alarmSound != "none") {
		var soundService = Components.classes["@mozilla.org/sound;1"].createInstance(Components.interfaces.nsISound);
		soundService.init();
		if (alarmSound == "beep")
		    soundService.beep();
		else if (alarmSound.substring(0, 6) == "chrome")
		    soundService.play(alarmSound);
		else
		    soundService.playSystemSound(alarmSound);
	    }
	}
	// see if we need to remove any alarms and kill them
	var a = 0;
	while (a < this.alarms.length) {
	    if (!this.alarms[a].active) {
		this.removeAlarm(this.alarms[a]);
		continue;
	    }
	    a++;
	}
    },

    addVanadielAlarmManagerListener: function(listener) {
	for (var l in this.mListeners) {
	    if (this.mListeners[l] == listener)
		return;
	}
	this.mListeners.push(listener);
	//dump("Vana'diel Alarm Manager: Added listener, up to " + this.mListeners.length + " listeners\n");
    },

    removeVanadielAlarmManagerListener: function(listener) {
	for (var l = 0; l < this.mListeners.length; l++) {
	    if (this.mListeners[l] == listener) {
		this.mListeners.splice(l, 1);
		//dump("Vana'diel Alarm Manager: Removed listener, down to " + this.mListeners.length + " listeners\n");
		return;
	    }
	}
	dump("Vana'diel Alarm Manager: Couldn't removed listener, listener does not exist in list\n");
    },

    QueryInterface: function (iid) {
        if (!iid.equals(Components.interfaces.nsIVanadielAlarmManager)
            && !iid.equals(Components.interfaces.nsISupports))
        {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    }
}

/**********************************************************
   XPCOM registration stuff, from
   http://books.mozdev.org/html/mozilla-chp-8-sect-2.html
   Modified to use an array of the several components we offer
 **********************************************************/

var Module = {
    firstTime: true,

    registerSelf: function (compMgr, fileSpec, location, type) {
        if (this.firstTime) {
            this.firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        compMgr =
	    compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
	for (var m in this.modules) {
	    compMgr.registerFactoryLocation(this.modules[m].mCID,
                                        this.modules[m].mName,
                                        this.modules[m].mProgID,
                                        fileSpec,
                                        location,
                                        type);
	}
    },

    getClassObject : function (compMgr, cid, iid) {
        if (!iid.equals(Components.interfaces.nsIFactory))
	    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
	for (var m in this.modules) {
	    if (cid.equals(this.modules[m].mCID))
		return this.modules[m].mFactory;
	}
	throw Components.results.NS_ERROR_NO_INTERFACE;
    },

    modules: {
	vanadielAlarm: {
	    mCID: Components.ID("{e8513bde-bad9-4172-a67f-8de617a55a57}"),
	    mProgID: "@xenoveritas.org/vanadiel-alarm;1",
	    mName: "Vana'diel Alarm",
	
	    mFactory: {
		createInstance: function (outer, iid) {
		    if (outer != null)
			throw Components.results.NS_ERROR_NO_AGGREGATION;
		    return (new VanadielAlarm()).QueryInterface(iid);
		}
	    },
	},
	vanadielAlarmManager: {
	    mCID: Components.ID("{e8513bde-bad9-4172-a67f-8de617a55a56}"),
	    mProgID: "@xenoveritas.org/vanadiel-alarm-manager;1",
	    mName: "Vana'diel Alarm Manager",
	
	    mFactory: {
		createInstance: function (outer, iid) {
		    if (outer != null)
			throw Components.results.NS_ERROR_NO_AGGREGATION;
		    return (new VanadielAlarmManager()).QueryInterface(iid);
		}
	    },
	}
    },

    canUnload: function(compMgr) {
        return true;
    }
}; // END Module

function NSGetModule(compMgr, fileSpec) { return Module; }
